package org.LiveGraph.dataCache;

import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;

import org.LiveGraph.dataFile.common.DataFormatException;
import org.LiveGraph.dataFile.read.DataStreamObserver;
import org.LiveGraph.dataFile.read.DataStreamReader;


/**
 * This reader will parse a data stream using {@link DataStreamReader} and store all
 * information in a data cache for further processing by the application.<br />
 * <br />
 * See {@link org.LiveGraph.dataFile.write.DataStreamWriter} for the details of the data file format. 
 * 
 * <p style="font-size:smaller;">This product includes software developed by the
 *    <strong>LiveGraph</strong> project and its contributors.<br />
 *    (<a href="http://www.live-graph.org" target="_blank">http://www.live-graph.org</a>)<br />
 *    Copyright (c) 2007 G. Paperin.<br />
 *    All rights reserved.
 * </p>
 * <p style="font-size:smaller;">File: DataStreamToCacheReader.java</p> 
 * <p style="font-size:smaller;">Redistribution and use in source and binary forms, with or
 *    without modification, are permitted provided that the following terms and conditions are met:
 * </p>
 * <p style="font-size:smaller;">1. Redistributions of source code must retain the above
 *    acknowledgement of the LiveGraph project and its web-site, the above copyright notice,
 *    this list of conditions and the following disclaimer.<br />
 *    2. Redistributions in binary form must reproduce the above acknowledgement of the
 *    LiveGraph project and its web-site, the above copyright notice, this list of conditions
 *    and the following disclaimer in the documentation and/or other materials provided with
 *    the distribution.<br />
 *    3. All advertising materials mentioning features or use of this software or any derived
 *    software must display the following acknowledgement:<br />
 *    <em>This product includes software developed by the LiveGraph project and its
 *    contributors.<br />(http://www.live-graph.org)</em><br />
 *    4. All advertising materials distributed in form of HTML pages or any other technology
 *    permitting active hyper-links that mention features or use of this software or any
 *    derived software must display the acknowledgment specified in condition 3 of this
 *    agreement, and in addition, include a visible and working hyper-link to the LiveGraph
 *    homepage (http://www.live-graph.org).
 * </p>
 * <p style="font-size:smaller;">THIS SOFTWARE IS PROVIDED &quot;AS IS&quot;, WITHOUT WARRANTY
 *    OF ANY KIND, EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND  NONINFRINGEMENT. IN NO EVENT SHALL
 *    THE AUTHORS, CONTRIBUTORS OR COPYRIGHT  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING  FROM, OUT OF OR
 *    IN CONNECTION WITH THE SOFTWARE OR THE USE OR  OTHER DEALINGS IN THE SOFTWARE.
 * </p>
 * 
 * @author Greg Paperin (<a href="http://www.paperin.org" target="_blank">http://www.paperin.org</a>)
 * @version {@value org.LiveGraph.LiveGraph#version}
 */
public class DataStreamToCacheReader implements Closeable, DataStreamObserver {

/**
 * Data stream reader.
 */
private DataStreamReader in = null;

/**
 * Cache for storage of extracted data.
 */
private DataCache cache = null;

/**
 * Creates a data reader on the specified stream.
 * 
 * @param is The stream from which to read.
 * @param cache The data cache into which to store the data. 
 */
public DataStreamToCacheReader(InputStream is, DataCache cache) {
	
	if (null == cache)
		throw new NullPointerException("Cannot use a null cache.");
	
	this.in = new DataStreamReader(is, this);	
	this.cache = cache;	
}

/**
 * @return The data cache used by this reader.
 */
public DataCache getDataCache() {
	return cache;
}

/**
 * Tells whether this reader's underlying data stream is ready to be read.
 * 
 * @return {@code true} if the next {@code readFromStream()} is guaranteed not to block for input,
 * {@code false} otherwise. Note that returning {@code false} does not guarantee that the next read will block.
 * @throws IOException If an I/O error occurs.
 */
public boolean ready() throws IOException {
	return in.ready();
}

/**
 * Closes the underlying data stream. Further reading is not possible after calling this method.
 * @throws IOException If an I/O error occurs.
 */
public void close() throws IOException {
	in.close();
}

/**
 * Reads as many data lines from the underlying stream as there are available, parses the lines and
 * stores the extracted information (if any) in this reader's data cache.
 *  
 * @return number of non-empty file lines read.
 * @throws IOException If an I/O error occurs.
 * @throws DataFormatException If the data stream contents do not conform with the expected data
 * stream format.
 * @see org.LiveGraph.dataFile.write.DataStreamWriter
 * @see org.LiveGraph.dataFile.read.DataStreamReader
 */
public int readFromStream() throws IOException, DataFormatException {
	
	int lines = in.readFromStream();
	return lines;	
}

/**
 * Used for callback by the {@link DataStreamReader}; does nothing.
 */
public void eventCommentLine(String line, DataStreamReader reader) {
	; // No action required.	
}

/**
 * Used for callback by the {@link DataStreamReader}; adds a dataset to the cache.
 */
public void eventDataLineRead(List<String> dataTokens, int datasetIndex, DataStreamReader reader) {
	
	List<Double> vals = DataStreamReader.convertTokensToDoubles(dataTokens);
	DataSet ds = new DataSet(vals, datasetIndex);
	cache.addDataSet(ds);	
}

/**
 * Used for callback by the {@link DataStreamReader}; adds a file info line to the cache.
 */
public void eventFileInfoLine(String info, DataStreamReader reader) {
	cache.addDataFileInfo(info);	
}

/**
 * Used for callback by the {@link DataStreamReader}; setts the data column labels in the cache.
 */
public void eventLabelsSet(List<String> labels, DataStreamReader reader) {
	List<String> uniqueLabels = DataStreamReader.createUniqueLabels(labels, true); 
	cache.resetLabels(uniqueLabels);	
}

/**
 * Used for callback by the {@link DataStreamReader}; does nothing.
 */
public void eventSeparatorSet(String separator, DataStreamReader reader) {
	; // No action required.	
}

}
